#! /usr/bin/env python
# ==============================================================================
# 
# LSST / StarDICE / ZTF / ... : Traveling CBP
# 
# Zolix Omni Lambda 2047i (200i family) - Python driver
#
# Authors: Kelian Sommer, Laurent Le Guillou, Johann Cohen-Tanugi
# Date: 2023-06-07
#
# ==============================================================================

import sys
import os, os.path
import time
import serial

# ==============================================================================

class ZolixError(Exception):
    pass

class ZolixTimeoutError(ZolixError):
    pass

# =========== Class Zolix controller ===========================================

class Zolix(object):
    """
    Driver for the Zolix Omni Lambda monochromator (USB-Serial)
    """

    # ==========================================================================
    
    ERROR_MEANINGS = {
        'E00': "This message indicates that the communication between the monochromator and the control program is not established, in this case, please send the contact instruction 'HELLO'.",
        'E01': "Communication error, send illegal control command or communication timeout.",
        'E02': "EEPROM error or the instrument has not yet set parameters, in this state, the monochromator only receives parameter setting commands.",
        'E03': "Grating parameter setting error, or not set.",
        'E04': "Wrong input of filter wheel parameters.",
        'E05': "The filter wheel command is enabled, but no equipment is actually installed or the equipment is faulty.",
        'E06': "Parameter settings other than the raster parameters are out of range. (Such as system parameters are not set incorrectly.)",
        'E07': "Stepping motor 3~9 issued a query position and move command before the positioning action is executed; or the zeroing action is executed, but the positioning switch is faulty. (If the system parameters are not set wrongly)",
        'E08': "The speed setting is out of the maximum running speed range.",
        'E09': "Double exit and double entrance switching commands are started, but no equipment is actually installed or the equipment is faulty.",
        'E10': "The encoder table coarse positioning switch is abnormal.",
        'E11': "Abnormal encoder table fine positioning switch.",
        'E12': "Auto slit command is enabled, but no equipment is actually installed or the equipment is faulty."
    }

    ERROR_CODES = ERROR_MEANINGS.keys()

    # For the following requests, the Zolix only sends 'OK\r' (or an error code)
    # 
    REQUEST_KEYWORDS_WITHOUT_EXTRA_DATA = [
        'HELLO',       # b'OK\r'
        'EXITPORT',    # b'OK\r'
        'FILTER'       # b'OK\r'
    ]
    
    REQUEST_KEYWORDS_WITH_EXTRA_DATA = [
        'SYSTEMINFO?', # b'SYSTEMINFO Zolix,Omni200,22040,20220901,V2.9\r'
        'EXITPORT?',   # b'EXITPORT 1\rOK\r'
        'POSITION?',   # b'POSITION 0\rOK\r'
        'GRATINGS?',   # b'GRATINGS\n1,600,300\n2,600,500\rOK\r'
        'GRATING?',    # b'GRATING 1\rOK\r'
        'FILTER?',     # b'FILTER 5\rOK\r'
        # 
        'MOVETO',      # b'399.999\rOK\r'
        'MOVE',        # b'399.999\rOK\r'
        'GRATING'      # b'0\rOK\r'
    ]

    REQUEST_KEYWORDS = REQUEST_KEYWORDS_WITHOUT_EXTRA_DATA + REQUEST_KEYWORDS_WITH_EXTRA_DATA

    # Exit port meaning (exit mirror positions)
    
    EXIT_PORT_FRONT = 0
    EXIT_PORT_SIDE = 1

    # ==========================================================================
    
    # ---------- Constructor ---------------------------------

    def __init__(self,
                 port = '/dev/serial/by-id/usb-Zolix_Omni-¦Ë2047i_OM222040-if00-port0',
                 serial_port = None, # to provide if serial port already created
                 debug = True):

        self.port = port
        self.serial_port = serial_port
        self.baudrate = 19200
        self.timeout = 0.5 # Non-blocking & non waiting mode
        self.repr_mode = 0
        self.parity = serial.PARITY_NONE
        self.bytesize = serial.EIGHTBITS
        self.stopbits = serial.STOPBITS_ONE
        self.serial_port = None
        self.request_EOM = '\r'
        self.answer_EOM = '\r'
        self.answer_line_sep = '\n'   # separator between lines for multilines answer
        self.answer_success = 'OK'
        self.answer_errors = self.ERROR_CODES

        # ---- debug mode
        self.debug = debug

        # ---- action timeout 
        self.default_exec_timeout = 25 # seconds  # to be tested with the GRATING/FILTER/EXITPORT cmds


    # ---------- Open the controller device ------------------

    def open(self):
        """
        Open the connexion (serial over USB) with the Zolix monochromator
        Check if there's something connected (echotest).
        """

        # if serial port has already been created
        # do not open it again

        if self.serial_port == None:
            if self.debug:
                print("Zolix: Opening port %s ..." % self.port, file=sys.stderr)

            self.serial_port = serial.Serial(port = self.port, 
                                             baudrate = self.baudrate,
                                             bytesize = self.bytesize, 
                                             parity = self.parity,
                                             stopbits = self.stopbits, 
                                             timeout = self.timeout)
        
            if ( (self.serial_port == None) or
                 not(self.serial_port.isOpen()) ):
                raise IOError("Zolix: Failed to open serial port %s" % self.port)

        # then the serial port is open (or was already open)
        
        self.serial_port.flushOutput()

        # Next, send the "HELLO" command

        status, extra = self.send("HELLO")

        if status != self.answer_success:
            raise IOError(("Zolix: monochromator is not properly responding on serial port %s") % 
                          self.port)

        if not(self.check_connection()):
            raise IOError(("Zolix: Wrong instrument signature on serial port %s") % self.port)
            
        if self.debug: 
            print(( "Zolix: Opening port %s done." % 
                                  self.port ), file=sys.stderr)
        

    # ---------- Close the controller device ----------------- 

    def close(self):
 
        if ( self.serial_port and
             self.serial_port.isOpen() ):
            self.serial_port.close()

    # ==========================================================================

    # ----------------- write command  ----------------------- 

    def write(self, request):
        """
        Send a command through the serial port. 
        """
        if not( self.serial_port and
                self.serial_port.isOpen() ):
            raise IOError("Zolix: Port %s not yet opened" % self.port)

        if self.debug: print("Zolix: Sending command [" + request + "]", file=sys.stderr)
        self.serial_port.write( (request + self.request_EOM).encode() )

    # ----------------- read command  ----------------------- 

    def read(self, serial_timeout = None):
        """
        Read the answer from the serial port, until we received a '\r'.
        Return it as a string.
        If <timeout> is specified, the function will wait
        for data with the specified timeout (instead of the default one). 
        Without any answer within <timeout>, an exception is raised.
        """
        
        if not( self.serial_port and
                self.serial_port.isOpen() ):
            raise IOError("Zolix: Port %s not yet opened" % self.port)

        if self.debug: print("Zolix: Reading serial port buffer", file=sys.stderr)

        if serial_timeout != None:
            self.serial_port.timeout = serial_timeout
            if self.debug: print("Zolix: Timeout specified: ", serial_timeout, file=sys.stderr)
            
        answer = self.serial_port.read_until(self.answer_EOM.encode())  # read until b'\r'
        
        # Restoring timeout to default one
        self.serial_port.timeout = self.timeout
        # 
        if self.debug: print("Zolix: Received [" + str(answer) + "]", file=sys.stderr)

        return answer.decode()

    # ----------------- Purge the serial port --------------- 

    def purge(self):
        """
        Purge the serial port to avoid framing errors.
        """
        self.serial_port.flushOutput()
        self.serial_port.flushInput()
        self.serial_port.readlines()

    # ==========================================================================

    # ---------- Send a command and get the answer -----------

    def send(self, keyword, args = [], request_timeout = None,
             check_syntax = True, raise_exception = True):
        """
        High level method to send a command to the Zolix monochromator ; 
        wait for the execution and the complete resulting answer. 
        Return the answer as a tuple: ('OK' or 'EXX', extra data (if any))

        @param keyword: the Zolix language keyword to send.
        @param args: the extra arguments (if any, [] otherwise)
        """
        # check if the request keyword is a legit one

        if check_syntax:
            if not(keyword in self.REQUEST_KEYWORDS):
                raise ValueError("Zolix: Invalid Zolix request (keyword is not legit)")
        
        # build the request

        request = keyword 
        if len(args) >= 1:
            request += ' '
            for arg in args[:-1]:
                request += arg + ','
            request += args[-1]
        
        # Now send it

        self.write(request)

        exec_timeout = self.default_exec_timeout
        if request_timeout != None:
            exec_timeout = request_timeout
            
        # Wait for the first part of the answer

        t_start = time.time()
        while True:
            delay = time.time() - t_start
            if delay > exec_timeout:
                raise ZolixTimeoutError("Zolix: Monochromator took too long to answer: no answer after %f s."
                                        % delay)

            answer_first_block = self.read().strip()
            if answer_first_block != '':
                break

        # Now analysing the first answer block ('XXXX\r')

        if keyword in self.REQUEST_KEYWORDS_WITHOUT_EXTRA_DATA:
            # in that case, the answer is complete
            return ( answer_first_block, [] )

        # For commands with extra data (self.REQUEST_KEYWORDS_WITH_EXTRA_DATA)

        if self.debug: print("Zolix: First answer block = ", answer_first_block, file = sys.stderr)

        if answer_first_block in self.answer_errors:
            error_code = answer_first_block
            if raise_exception:
                raise ZolixError( "Zolix: " +
                                  error_code + ': ' + self.ERROR_MEANINGS[error_code] )
            # otherwise return the error code
            return ( error_code, [] )
        
        # Otherwise, wait for the [second] part of the answer
        
        t_start = time.time()
        while True:
            delay = time.time() - t_start
            if delay > exec_timeout:
                raise ZolixTimeoutError("Zolix: Monochromator took too long to answer: no answer after %f s."
                                        % delay)

            answer_second_block = self.read().strip()
            if answer_second_block != '':
                break

        if self.debug:
            print("Zolix: Second answer block = ", answer_second_block, file = sys.stderr)

        if answer_second_block in self.answer_errors:
            error_code = answer_second_block
            if raise_exception:
                raise ZolixError( "Zolix: " +
                                  error_code + ': ' + self.ERROR_MEANINGS[error_code] )
        
        # Then split the first answer block on '\n' as the first block contains data
        # and the second block contains the OK / EXX code.

        extra_parts = answer_first_block.split(self.answer_line_sep)

        return ( answer_second_block, extra_parts )

    # ==========================================================================
    
    # ---------- Get the Zolix monochromator signature -------

    def get_system_info(self, raise_exception = True):
        """
        Return the Zolix model, serial number, etc
        """
        status, extra = self.send("SYSTEMINFO?", raise_exception = raise_exception)

        # S.write(b'SYSTEMINFO?\r'); S.read_until(b'\r')
        # b'SYSTEMINFO Zolix,Omni200,22040,20220901,V2.9\r'

        # check that extra as at least one element ???
        
        signature = extra[0].replace('SYSTEMINFO ', '').split(',')
        
        return signature

    # --------------------------------------------------------
    
    system_info = property(get_system_info, doc="Zolix instrument signature")

    # ---------- Get and check the instrument signature --------

    def check_connection(self):
        """
        Check the instrument signature. Return True/False.
        """
        status, extra = self.send("SYSTEMINFO?", raise_exception = False)

        if status != self.answer_success:
            return False

        if len(extra) != 1:
            return False

        # S.write(b'SYSTEMINFO?\r'); S.read_until(b'\r')
        # b'SYSTEMINFO Zolix,Omni200,22040,20220901,V2.9\r'

        pattern = 'SYSTEMINFO Zolix'
        if not(pattern in extra[0]):
            return False

        return True

    # ==========================================================================
    
    # ---------- Get the current wavelength --------------------

    def get_wavelength(self):
        """
        Return the current wavelength [nm].
        """
        status, extra = self.send("POSITION?", raise_exception = True)

        if len(extra) != 1:
            raise ZolixError("Zolix: invalid answer format [POSITION?].")

        try:
            value_str = extra[0].replace('POSITION ', '')
            wl = float(value_str)
        except ValueError:
            raise ZolixError("Zolix: invalid wavelength [POSITION?]")
            
        return wl

    # ---------- Set the current wavelength --------------------

    def set_wavelength(self, wl):
        """
        Set the current wavelength (absolute move) [nm].
        Return the effective wavelength reached.
        """
        status, extra = self.send("MOVETO", ['%f' % wl], raise_exception = True)

        if len(extra) != 1:
            raise ZolixError("Zolix: invalid answer format [MOVETO].")

        try:
            value_str = extra[0]
            value = float(value_str) # effective wavelength
        except ValueError:
            raise ZolixError("Zolix: invalid returned wavelength value [MOVETO]")
            
        return value


    # ---------- Get/Set the wavelength [as a property] --------

    wavelength = property(get_wavelength, set_wavelength, doc="current wavelength")

    # ==========================================================================

    def appropriate_filter(self, wl):
        """
        Return the appropriate filter wheel position when setting the 
        monochromator to wavelength <wl>, in order to cut the second order 
        contamination.
        This depends of the actual filters installed in the filter wheels.
        """
        if wl < 360.0:
            return 1

        if wl < 450.0:
            return 2

        if wl < 650.0:
            return 3

        if wl < 800.0:
            return 4

        return 5
        
    
    # ---------- High level: set wavelength and appropriate filter

    def set_wavelength_and_filter(self, wl):
        """
        High level 'smart' method to set the monochromator 
        wavelength and select the appropriate filter in the
        filter wheel to reduce contamination from the
        grating 2nd and higher orders.
        """

        app_filt = self.appropriate_filter(wl)
        self.set_filter(app_filt)
        self.purge() # necessary to avoid weird crash happening sometimes
        self.set_wavelength(wl)

    # ==========================================================================

    # ---------- Get the current filter wheel position ---------

    def get_filter(self):
        """
        Return the current filter wheel position.
        """
        status, extra = self.send("FILTER?", raise_exception = True)

        if len(extra) != 1:
            raise ZolixError("Zolix: invalid answer format [FILTER?].")

        try:
            value_str = extra[0].replace('FILTER ', '')
            value = int(value_str)
        except ValueError:
            raise ZolixError("Zolix: invalid filter wheel position value [FILTER?]")
            
        return value

    # ---------- Set the current filter wheel position ---------

    def set_filter(self, filt):
        """
        Set the current filter wheel position.
        """
        status, extra = self.send("FILTER", ['%d' % filt], raise_exception = True)

    # ---------- Get/Set the filter wheel position [property] --

    filt = property(get_filter, set_filter, doc="current filter wheel position")
    
    # ==========================================================================
    
    # ---------- Get the current exit port (mirror position) ---

    def get_exit_port(self):
        """
        Return the current exit port (exit mirror position).
        """
        status, extra = self.send("EXITPORT?", raise_exception = True)

        if len(extra) != 1:
            raise ZolixError("Zolix: invalid answer format [EXITPORT?].")

        try:
            value_s = extra[0].replace('EXITPORT ', '')
            value = int(value_s)
        except ValueError:
            raise ZolixError("Zolix: invalid exit port value [EXITPORT?]")
            
        return value

    # ---------- Set the current exit port (mirror position) ---

    def set_exit_port(self, position):
        """
        Set the current exit port by moving the exit mirror.
        """
        status, extra = self.send("EXITPORT", ['%d' % position], raise_exception = True)

    # ---------- Get/Set the wavelength [as a property] --------

    exit_port = property(get_exit_port, set_exit_port, doc="current exit port (exit mirror position)")
    
    # ==========================================================================

    # ---------- Get the current grating -----------------------

    def get_grating(self):
        """
        Return the current grating.
        """
        status, extra = self.send("GRATING?", raise_exception = True)

        if len(extra) != 1:
            raise ZolixError("Zolix: invalid answer format [GRATING?].")

        try:
            value_str = extra[0].replace('GRATING ', '')
            value = int(value_str)
        except ValueError:
            raise ZolixError("Zolix: invalid grating value [GRATING?]")
            
        return value

    # ---------- Set the current grating -----------------------

    def set_grating(self, g):
        """
        Set the current grating.
        """
        status, extra = self.send("GRATING", ['%d' % g], raise_exception = True)

    # ---------- Get/Set the grating [property] ----------------

    grating = property(get_grating, set_grating, doc="current grating")
   
    # ---------- Retrieve info about the installed gratings ----

    def get_gratings(self):
        """
        Retrieve the properties of the installed gratings.
        """
        status, extra = self.send("GRATINGS?", raise_exception = True)

        if len(extra) < 1:
            raise ZolixError("Zolix: invalid answer format [GRATINGS?].")

        # at this point, extra should look like ['GRATINGS', '1,600,300', '2,600,500', ... ]

        # it may be useful to properly extract the gratings parameters...
        # TO BE DONE FOR A FURTHER VERSION

        return extra[1:]

    # ---------- Get the properties of the installed gratings --

    gratings = property(get_gratings, doc="current grating")

    
    # ==========================================================================
    

# ===============================================================================
